# Brainstorming for an ModelEstimator Class, which takes a time series and provides methods for estimating a Model
# describing the time evolution
# this class needs to house a subclass "Model", which in turn needs to create a library of possible terms in the right
# hand side

